"""
This module provides the FastMockDecorator class for adding mock data attributes to functions.
"""
from fastmock.model import MockData


class FastMockDecorator:
    """
    A decorator class for adding mock data attributes to functions.

    This class allows you to define default mock data and update it when applying the decorator
    to functions.
    The mock data is stored as an attribute on the decorated function.

    Attributes:
        attribute_name (str): The name of the attribute that will be added to the decorated
        function to store the mock data.

    Args:
        **kwargs: Keyword arguments that are used to initialize the default mock data.

    Methods:
        __call__(func=None, **kwargs):
            If `func` is None, updates the default mock data with the provided kwargs and
            returns a new instance of FastMockDecorator.
            If `func` is provided, updates the default mock data with the provided kwargs,
            sets it as an attribute on the function, and returns the function.
    """

    attribute_name = "mock_data"

    def __init__(self, **kwargs):
        """
        Initializes the FastMockDecorator with default mock data.

        Args:
            **kwargs: Keyword arguments to initialize the default mock data.
        """
        self.default_data = MockData(**kwargs)

    def __call__(self, func=None, **kwargs):
        """
        Updates the default mock data and decorates the given function with the updated mock data.

        If `func` is None, returns a new instance of FastMockDecorator with the updated mock data.
        If `func` is provided, sets the updated mock data as an attribute on the function
        and returns the function.

        Args:
            func (callable, optional): The function to be decorated with the mock data.
            **kwargs: Keyword arguments to update the default mock data.

        Returns:
            If `func` is None, returns a new instance of FastMockDecorator
            with the updated mock data.
            If `func` is provided, returns the decorated function with the updated mock data.
        """
        if func is None:
            updated_data = self.default_data.copy(update=kwargs)
            return self.__class__(**updated_data.dict())

        updated_data = self.default_data.copy(update=kwargs)
        setattr(func, self.attribute_name, updated_data)
        return func



import pytest
from tests.clients import (decorator_client, decorator_initialization_client,
                           middleware_client, header_client)


@pytest.mark.parametrize("get_client", [
    middleware_client.get_client,
    decorator_client.get_client,
    decorator_initialization_client.get_client,
    header_client.get_client
])
@pytest.mark.parametrize("activate", [True, False])
def test_fastmock_activation(get_client, activate: bool):
    client = get_client(activate=activate, type="example")
    response = client.get("/device")
    assert response.status_code == 200

    if activate:
        assert response.json() == {
            "device_uuid": "DEVX000001",
            "localisation": {
                "latitude": 35.6582,
                "longitude": 139.8752
            },
            "deployment_date": "2024-03-14",
            "owner": "tanguy.pemeja@gmail.com"
        }
    else:
        assert response.json() == {'msg': 'Device'}


@pytest.mark.parametrize("get_client", [
    middleware_client.get_client,
    decorator_client.get_client,
    decorator_initialization_client.get_client,
    header_client.get_client
])
@pytest.mark.parametrize("value_type", ["dict", "list", "device", "str", "int"])
def test_fastmock_activation_value(get_client, value_type: str):
    client = get_client(activate=False)
    response = client.get(f"/{value_type}")
    assert response.status_code == 200
    assert response.json().get("msg", False)



from fastapi import FastAPI, status
from fastapi.testclient import TestClient
from fastmock.model import MockData

from fastmock.middleware import FastMockMiddleware
from tests.device import Device, DeviceNotFoundResponse


def get_client(**kwargs) -> TestClient:
    # Recreate the app to reset the middleware for each test
    test_app = FastAPI()
    test_app.add_middleware(FastMockMiddleware, mock_data=MockData(**kwargs))

    @test_app.get("/list",
                  status_code=status.HTTP_200_OK,
                  responses={
                      status.HTTP_200_OK: {"model": list[Device]},
                      status.HTTP_404_NOT_FOUND: {"model": DeviceNotFoundResponse}
                  })
    async def get_device_list():
        return {"msg": "Device list"}

    @test_app.get("/dict",
                  status_code=status.HTTP_200_OK,
                  responses={
                      status.HTTP_200_OK: {"model": dict[str, Device]},
                      status.HTTP_404_NOT_FOUND: {"model": DeviceNotFoundResponse}
                  })
    async def get_device():
        return {"msg": "Device dict"}

    @test_app.get("/device",
                  status_code=status.HTTP_200_OK,
                  responses={
                      status.HTTP_200_OK: {"model": Device},
                      status.HTTP_404_NOT_FOUND: {"model": DeviceNotFoundResponse}
                  })
    async def get_device():
        return {"msg": "Device"}

    @test_app.get("/str",
                  status_code=status.HTTP_200_OK,
                  responses={
                      status.HTTP_200_OK: {"model": str},
                      status.HTTP_404_NOT_FOUND: {"model": DeviceNotFoundResponse}
                  })
    async def get_str():
        return {"msg": "Str"}

    @test_app.get("/int",
                  status_code=status.HTTP_200_OK,
                  responses={
                      status.HTTP_200_OK: {"model": int},
                      status.HTTP_404_NOT_FOUND: {"model": DeviceNotFoundResponse}
                  })
    async def get_int():
        return {"msg": "Int"}

    return TestClient(test_app)



import pytest

from tests.clients import (decorator_client, decorator_initialization_client,
                           middleware_client, header_client)

swagger_json = ""

@pytest.mark.parametrize("get_client", [
    middleware_client.get_client,
    decorator_client.get_client,
    decorator_initialization_client.get_client,
    header_client.get_client
])
def test_fastmock_openapi_working(get_client):
    response = get_client().get("/openapi.json")
    assert response.status_code == 200
    assert (response.json() ==
            {
                'openapi': '3.1.0',
                'info': {
                    'title': 'FastAPI',
                    'version': '0.1.0'
                },
                'paths': {
                    '/list': {
                        'get': {
                            'summary': 'Get Device List',
                            'operationId': 'get_device_list_list_get',
                            'responses': {
                                '200': {
                                    'description': 'Successful Response',
                                    'content': {
                                        'application/json': {
                                            'schema': {
                                                'items': {
                                                    '$ref': '#/components/schemas/Device'
                                                },
                                                'type': 'array',
                                                'title': 'Response 200 Get Device List List Get'
                                            }
                                        }
                                    }
                                },
                                '404': {
                                    'description': 'Not Found',
                                    'content': {
                                        'application/json': {
                                            'schema': {
                                                '$ref': '#/components/schemas/DeviceNotFoundResponse'
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                    '/dict': {
                        'get': {
                            'summary': 'Get Device',
                            'operationId': 'get_device_dict_get',
                            'responses': {
                                '200': {
                                    'description': 'Successful Response',
                                    'content': {
                                        'application/json': {
                                            'schema': {
                                                'additionalProperties': {
                                                    '$ref': '#/components/schemas/Device'
                                                },
                                                'type': 'object',
                                                'title': 'Response 200 Get Device Dict Get'
                                            }
                                        }
                                    }
                                },
                                '404': {
                                    'description': 'Not Found',
                                    'content': {
                                        'application/json': {
                                            'schema': {
                                                '$ref': '#/components/schemas/DeviceNotFoundResponse'
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                    '/device': {
                        'get': {
                            'summary': 'Get Device',
                            'operationId': 'get_device_device_get',
                            'responses': {
                                '200': {
                                    'description': 'Successful Response',
                                    'content': {
                                        'application/json': {
                                            'schema': {
                                                '$ref': '#/components/schemas/Device'
                                            }
                                        }
                                    }
                                },
                                '404': {
                                    'description': 'Not Found',
                                    'content': {
                                        'application/json': {
                                            'schema': {
                                                '$ref': '#/components/schemas/DeviceNotFoundResponse'
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                    '/str': {
                        'get': {
                            'summary': 'Get Str',
                            'operationId': 'get_str_str_get',
                            'responses': {
                                '200': {
                                    'description': 'Successful Response',
                                    'content': {
                                        'application/json': {
                                            'schema': {
                                                'type': 'string',
                                                'title': 'Response 200 Get Str Str Get'
                                            }
                                        }
                                    }
                                },
                                '404': {
                                    'description': 'Not Found',
                                    'content': {
                                        'application/json': {
                                            'schema': {
                                                '$ref': '#/components/schemas/DeviceNotFoundResponse'
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                    '/int': {
                        'get': {
                            'summary': 'Get Int',
                            'operationId': 'get_int_int_get',
                            'responses': {
                                '200': {
                                    'description': 'Successful Response',
                                    'content': {
                                        'application/json': {
                                            'schema': {
                                                'type': 'integer',
                                                'title': 'Response 200 Get Int Int Get'
                                            }
                                        }
                                    }
                                },
                                '404': {
                                    'description': 'Not Found',
                                    'content': {
                                        'application/json': {
                                            'schema': {
                                                '$ref': '#/components/schemas/DeviceNotFoundResponse'
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                },
                'components': {
                    'schemas': {
                        'Coordinate': {
                            'properties': {
                                'latitude': {
                                    'type': 'number',
                                    'title': 'Latitude',
                                    'description': 'Latitude in decimal degrees. Must be between -90 and 90.'
                                },
                                'longitude': {
                                    'type': 'number',
                                    'title': 'Longitude',
                                    'description': 'Longitude in decimal degrees. Must be between -180 and 180.'
                                }
                            },
                            'type': 'object',
                            'required': [
                                'latitude',
                                'longitude'
                            ],
                            'title': 'Coordinate',
                            'description': 'Represents a geographical coordinate with latitude and longitude.\n\nAttributes:\n- latitude (float): The latitude in decimal degrees, must be between -90 and 90.\n- longitude (float): The longitude in decimal degrees, must be between -180 and 180.',
                            'example': {
                                'latitude': 35.6582,
                                'longitude': 139.8752
                            }
                        },
                        'Device': {
                            'properties': {
                                'device_uuid': {
                                    'type': 'string',
                                    'title': 'Device Uuid',
                                    'description': 'The UUID of the device. It should start with the prefix (DEV), a single uppercase letter, and six digits (e.g., DEVX000001)'
                                },
                                'localisation': {
                                    'allOf': [
                                        {
                                            '$ref': '#/components/schemas/Coordinate'
                                        }
                                    ],
                                    'description': 'The geographical location of the device as latitude and longitude coordinates'
                                },
                                'deployment_date': {
                                    'type': 'string',
                                    'format': 'date',
                                    'title': 'Deployment Date',
                                    'description': 'The deployment date of the device in the format YYYY-MM-DD'
                                },
                                'owner': {
                                    'type': 'string',
                                    'format': 'email',
                                    'title': 'Owner',
                                    'description': 'The email address of the owner of the device'
                                }
                            },
                            'type': 'object',
                            'required': [
                                'device_uuid',
                                'owner'
                            ],
                            'title': 'Device',
                            'description': 'Represents a device with unique identifiers and location information.\n\nAttributes:\n- device_uuid (str): The UUID of the device. It should start with the prefix (DEV),\n  followed by a single uppercase letter and six digits (e.g., DEVX000001).\n- localisation (Coordinate): Geographical location of the device as coordinates.\n- deployment_date (date): The deployment date of the device in the format YYYY-MM-DD.\n- owner (EmailStr): The email address of the owner of the device.',
                            'example': {
                                'deployment_date': '2024-03-14',
                                'device_uuid': 'DEVX000001',
                                'localisation': {
                                    'latitude': 35.6582,
                                    'longitude': 139.8752
                                },
                                'owner': 'tanguy.pemeja@gmail.com'
                            }
                        },
                        'DeviceNotFoundResponse': {
                            'properties': {
                                'message': {
                                    'type': 'string',
                                    'title': 'Message',
                                    'default': 'Device not found'
                                }
                            },
                            'type': 'object',
                            'title': 'DeviceNotFoundResponse',
                            'description': 'Represents a response indicating that the device was not found.'
                        }
                    }
                }
            })



from typing import Callable

from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware

from fastmock.request_response import get_matched_route, get_response
from fastmock.model import MockData

from fastmock.tools import get_data_from_decorator_route, get_data_from_header


class FastMockMiddleware(BaseHTTPMiddleware):
    """
    Middleware for FastAPI to mock responses based on predefined data.

    This middleware intercepts requests and provides mock responses based on
    specified rules and data. It allows for flexible retrieval and merging of
    mock data from various sources.

    Attributes:
        retrieve_data_function_list (list[Callable[[Request], dict]]): List of functions
            to retrieve additional data from the request, ordered from least to most
            important.
        mock_data (MockData): The base mock data configuration.
    """

    def __init__(
            self,
            app,
            mock_data: MockData = MockData(),
            retrieve_data_function_list: list[Callable[[Request], dict]] = None
    ):
        """
        Initializes the FastMockMiddleware.

        Args:
            app: The FastAPI application instance.
            mock_data (MockData): The base mock data configuration. Defaults to an empty MockData instance.
            retrieve_data_function_list (list[Callable[[Request], dict]]): List of functions
                to retrieve additional data from the request, ordered from least to most
                important function. Defaults to [get_data_from_decorator_route, get_data_from_header].
        """
        super().__init__(app)

        if retrieve_data_function_list is None:
            retrieve_data_function_list = [get_data_from_decorator_route, get_data_from_header]

        self.retrieve_data_function_list = retrieve_data_function_list
        self.mock_data = mock_data

    def get_mock_data(self, request: Request) -> MockData:
        """
        Merges base mock data with additional data retrieved from the request.

        Args:
            request (Request): The incoming HTTP request.

        Returns:
            MockData: The merged mock data.
        """
        merged_data = self.mock_data.dict()
        for retrieve_data_function in self.retrieve_data_function_list:
            merged_data.update(retrieve_data_function(request))

        return MockData(**merged_data)

    async def dispatch(self, request, call_next):
        """
        Intercepts the request and returns a mock response if applicable.

        Args:
            request (Request): The incoming HTTP request.
            call_next (Callable): The next callable in the middleware chain.

        Returns:
            Response: The HTTP response.
        """
        mock_data = self.get_mock_data(request)

        # Check if the route matches and if mocking is activated
        if get_matched_route(request) is None or not mock_data.activate:
            response = await call_next(request)
            return response

        # Get the mock response based on the request and mock data
        route_response = get_response(request, mock_data)

        return route_response



from fastapi import FastAPI, status
from fastapi.testclient import TestClient

from fastmock.middleware import FastMockMiddleware
from tests.device import Device, DeviceNotFoundResponse
from fastmock.decorator import FastMockDecorator
from fastmock.tools import get_data_from_decorator_route, get_data_from_header


def get_client(retrieve_data_function_list: list | None) -> TestClient:
    # Recreate the app to reset the middleware for each test
    test_app = FastAPI()
    test_app.add_middleware(FastMockMiddleware,
                            retrieve_data_function_list=retrieve_data_function_list)

    mock = FastMockDecorator()

    @mock(activate=False, type="example")
    @test_app.get("/",
                  status_code=status.HTTP_200_OK,
                  responses={
                      status.HTTP_200_OK: {"model": list[Device]},
                      status.HTTP_404_NOT_FOUND: {"model": DeviceNotFoundResponse}
                  })
    async def get_device_list():
        return {"msg": "Device list"}

    return TestClient(test_app)


def test_fastmock_default_data_order():
    client = get_client(retrieve_data_function_list=None)
    response = client.get("/", headers={'X-FASTMOCK-ACTIVATE': 'True'})
    assert response.json() == [
        {
            'deployment_date': '2024-03-14',
            'device_uuid': 'DEVX000001',
            'localisation':
                {'latitude': 35.6582,
                 'longitude': 139.8752
                 },
            'owner': 'tanguy.pemeja@gmail.com'
        },
        {
            'deployment_date': '2024-03-14',
            'device_uuid': 'DEVX000001',
            'localisation':
                {'latitude': 35.6582,
                 'longitude': 139.8752
                 },
            'owner': 'tanguy.pemeja@gmail.com'
        }]


def test_fastmock_data_order():
    client = get_client(
        retrieve_data_function_list=[get_data_from_decorator_route, get_data_from_header])
    response = client.get("/", headers={'X-FASTMOCK-ACTIVATE': 'True'})
    assert response.json() == [
        {
            'deployment_date': '2024-03-14',
            'device_uuid': 'DEVX000001',
            'localisation':
                {'latitude': 35.6582,
                 'longitude': 139.8752
                 },
            'owner': 'tanguy.pemeja@gmail.com'
        },
        {
            'deployment_date': '2024-03-14',
            'device_uuid': 'DEVX000001',
            'localisation':
                {'latitude': 35.6582,
                 'longitude': 139.8752
                 },
            'owner': 'tanguy.pemeja@gmail.com'
        }]


def test_fastmock_data_reorder():
    client = get_client(
        retrieve_data_function_list=[get_data_from_header, get_data_from_decorator_route])
    response = client.get("/", headers={'X-FASTMOCK-ACTIVATE': 'True'})
    assert response.json() == {'msg': 'Device list'}



import re
from types import GenericAlias
from typing import Any, Mapping, Optional, Sequence, get_args, get_origin, Type

from fastapi import Request
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from fastapi.routing import APIRoute
from polyfactory import BaseFactory

from fastmock.model import MockData


def get_matched_route(request: Request) -> APIRoute | None:
    """
    Matches the request to a defined APIRoute in the FastAPI application.

    Args:
        request (Request): The incoming HTTP request.

    Returns:
        APIRoute | None: The matched route or None if no match is found.
    """
    for route in request["app"].routes:
        if isinstance(route, APIRoute):
            # Convert route path to regex pattern
            route_pattern = re.sub(r"{\w+}", r"[^/]+", route.path)
            route_pattern = f"^{route_pattern}$"

            # Check if requested path matches the regex pattern of the route
            if re.match(route_pattern, request["path"]) and request["method"] in route.methods:
                return route

    return None


def get_model_response(model, mock_data: MockData):
    """
    Generates a response based on the provided model and mock data.

    Args:
        model: The model class to generate the response for.
        mock_data (MockData): The mock data configuration.

    Returns:
        Any: The generated model response.
    """
    if isinstance(model, GenericAlias) and issubclass(get_origin(model), Sequence):
        return [get_model_factory(get_args(model)[0], mock_data)
                for _ in range(mock_data.element_size)]

    if isinstance(model, GenericAlias) and issubclass(get_origin(model), Mapping):
        key_type, value_type = get_args(model)
        keys = BaseFactory.__faker__.pylist(
            nb_elements=mock_data.element_size,
            variable_nb_elements=False,
            value_types=[key_type]
        )

        model_response = {}
        for key in keys:
            model_response[key] = get_model_factory(value_type, mock_data)

        return model_response

    return get_model_factory(model, mock_data)


def get_model_factory(model, mock_data: MockData):
    """
    Creates a factory for the provided model and generates mock data.

    Args:
        model: The model class to generate the factory for.
        mock_data (MockData): The mock data configuration.

    Returns:
        Any: The generated model instance.
    """
    if mock_data.type == mock_data.type.example:
        model_example = (getattr(model, "model_config", {})
                         .get("json_schema_extra", {}).get("example", None))

        if model_example:
            return model_example

        raise Exception("Mock using model example but no example found for the API")

    factory_class = get_mock_factory_class(model)

    if factory_class:
        factory = factory_class.create_factory(
            model=model,
            __use_defaults__=mock_data.type == mock_data.type.default
        )
        return factory.build()

    provider = BaseFactory.get_provider_map().get(model, None)

    if provider:
        return provider()

    raise ValueError(f'Cannot mock {model.response_model.__name__}')


def get_mock_factory_class(response_model: Any) -> Optional[Type[BaseFactory]]:
    """
    Retrieves the appropriate factory class for the provided response model.

    Args:
        response_model (Any): The response model class.

    Returns:
        Optional[Type[BaseFactory]]: The factory class if found, else None.
    """
    for factory in BaseFactory.__subclasses__():
        if factory.is_supported_type(response_model):
            return factory

    return None


def get_response(request: Request, mock_data: MockData):
    """
    Generates a JSON response based on the matched route and mock data.

    Args:
        request (Request): The incoming HTTP request.
        mock_data (MockData): The mock data configuration.

    Returns:
        JSONResponse: The generated JSON response.
    """
    api_route = get_matched_route(request)

    if mock_data.response_status_code is None:
        status_code = api_route.status_code
    else:
        status_code = mock_data.response_status_code

    if status_code not in api_route.responses:
        raise Exception("Mock status code not defined in API declaration")
    api_response = api_route.responses[status_code]

    response_model = get_model_response(api_response.get("model"), mock_data)

    return JSONResponse(status_code=status_code,
                        content=jsonable_encoder(response_model))



import pytest
from tests.clients import (decorator_client, decorator_initialization_client,
                           middleware_client, header_client)


@pytest.mark.parametrize("get_client", [
    middleware_client.get_client,
    decorator_client.get_client,
    decorator_initialization_client.get_client
])
@pytest.mark.parametrize("size", [1, 2, 5])
def test_fastmock_list_size_response(get_client, size: int):
    client = get_client(element_size=size, type='example')
    response = client.get("/list")
    assert response.status_code == 200
    assert response.json() == [{
        "device_uuid": "DEVX000001",
        "localisation": {
            "latitude": 35.6582,
            "longitude": 139.8752
        },
        "deployment_date": "2024-03-14",
        "owner": "tanguy.pemeja@gmail.com"
    }] * size


@pytest.mark.parametrize("get_client", [
    middleware_client.get_client,
    decorator_client.get_client,
    decorator_initialization_client.get_client,
    header_client.get_client
])
@pytest.mark.parametrize("size", [1, 2, 5])
def test_fastmock_dict_size_response(get_client, size: int):
    client = get_client(element_size=size, type='example')
    response = client.get("/dict")
    assert response.status_code == 200
    assert list(response.json().values()) == [{
        "device_uuid": "DEVX000001",
        "localisation": {
            "latitude": 35.6582,
            "longitude": 139.8752
        },
        "deployment_date": "2024-03-14",
        "owner": "tanguy.pemeja@gmail.com"
    }] * size



from fastapi import FastAPI, status
from fastapi.testclient import TestClient

from fastmock.middleware import FastMockMiddleware
from tests.device import Device, DeviceNotFoundResponse
from fastmock.decorator import FastMockDecorator
from fastmock.model import MockData

def get_client() -> TestClient:
    # Recreate the app to reset the middleware for each test
    test_app = FastAPI()
    test_app.add_middleware(FastMockMiddleware, mock_data=MockData(activate=False))

    mock = FastMockDecorator()

    @mock
    @test_app.get("/",
                  status_code=status.HTTP_200_OK,
                  responses={
                      status.HTTP_200_OK: {"model": list[Device]},
                      status.HTTP_404_NOT_FOUND: {"model": DeviceNotFoundResponse}
                  })
    async def get_device_list():
        return {"msg": "Device list"}

    return TestClient(test_app)


def test_fastmock_decorator_work_without_parenthesis():
    client = get_client()
    response = client.get("/")
    assert response.status_code == 200
    assert response.json() != {"msg": "Device list"}



from enum import Enum
from pydantic import BaseModel


class GenerationTypeEnum(str, Enum):
    """
    Enumeration for the type of generation for mock data.

    Attributes:
        default (str): Represents default data.
        example (str): Represents example data.
        generated (str): Represents data generated dynamically.
    """
    default = 'default'
    example = 'example'
    generated = 'generated'


class MockData(BaseModel):
    """
    Model for defining mock data configurations.

    Attributes:
        activate (bool): Flag to activate or deactivate mock responses. Defaults to True.
        element_size (int): Number of elements to be included in the mock response. Defaults to 2.
        type (GenerationTypeEnum): The type of data generation. Defaults to GenerationTypeEnum.default.
        response_status_code (int | None): HTTP status code for the mock response. If None, a default status code is used.
    """
    activate: bool = True
    element_size: int = 2
    type: GenerationTypeEnum = GenerationTypeEnum.default
    response_status_code: int | None = None



import pytest
from tests.clients import (decorator_client, decorator_initialization_client,
                           middleware_client, header_client)


@pytest.mark.parametrize("get_client", [
    middleware_client.get_client,
    decorator_client.get_client,
    decorator_initialization_client.get_client,
    header_client.get_client
])
@pytest.mark.parametrize("value_type", ["dict", "list", "device", "str"])
def test_fastmock_response_type(get_client, value_type: str):
    try:
        client = get_client(type='example')
        response = client.get(f"/{value_type}")
        assert response.status_code == 200

        match value_type:
            case "dict":
                assert isinstance(response.json(), dict)
                for value in response.json().values():
                    assert value == {
                        "device_uuid": "DEVX000001",
                        "localisation": {
                            "latitude": 35.6582,
                            "longitude": 139.8752
                        },
                        "deployment_date": "2024-03-14",
                        "owner": "tanguy.pemeja@gmail.com"
                    }
            case "list":
                assert isinstance(response.json(), list)
                for value in response.json():
                    assert value == {
                        "device_uuid": "DEVX000001",
                        "localisation": {
                            "latitude": 35.6582,
                            "longitude": 139.8752
                        },
                        "deployment_date": "2024-03-14",
                        "owner": "tanguy.pemeja@gmail.com"
                    }
            case "device":
                assert response.json() == {
                    "device_uuid": "DEVX000001",
                    "localisation": {
                        "latitude": 35.6582,
                        "longitude": 139.8752
                    },
                    "deployment_date": "2024-03-14",
                    "owner": "tanguy.pemeja@gmail.com"
                }
            case _:
                assert False

    except Exception as e:
        match value_type:
            case "str":
                assert str(e) == "Mock using model example but no example found for the API"
            case _:
                assert False



import re
from fastapi import Request
from fastmock.decorator import FastMockDecorator
from fastmock.request_response import get_matched_route


def get_data_from_decorator_route(request: Request) -> dict:
    """
    Retrieves mock data from a route's decorator if it exists.

    Args:
        request (Request): The incoming HTTP request.

    Returns:
        dict: The mock data retrieved from the route's decorator, or an empty dictionary if not found.
    """
    route = get_matched_route(request)
    if route is not None and hasattr(route.endpoint, FastMockDecorator.attribute_name):
        return getattr(route.endpoint, FastMockDecorator.attribute_name).dict()

    return {}


def get_data_from_header(request: Request, header_prefix: str = "X-FASTMOCK-") -> dict:
    """
    Retrieves mock data from request headers that match a specified prefix.

    Args:
        request (Request): The incoming HTTP request.
        header_prefix (str): The prefix to match headers against. Defaults to "X-FASTMOCK-".

    Returns:
        dict: The mock data retrieved from the headers.
    """
    header_data = {}
    # Create a regex pattern to match the prefix case-insensitively
    pattern = re.compile(re.escape(header_prefix), re.IGNORECASE)

    for header_name, header_value in request.headers.items():
        match = pattern.match(header_name)
        if match:
            # Extract the part after the prefix
            extracted_name = header_name[match.end():].lower().replace("-", "_")
            header_data[extracted_name] = header_value

    return header_data



import pytest
from tests.clients import (decorator_client, decorator_initialization_client,
                           middleware_client, header_client)


@pytest.mark.parametrize("get_client", [
    middleware_client.get_client,
    decorator_client.get_client,
    decorator_initialization_client.get_client,
    header_client.get_client
])
@pytest.mark.parametrize("response_status_code", [200, 404, 409])
def test_fastmock_status_code(get_client, response_status_code: int):
    try:
        client = get_client(response_status_code=response_status_code, type='example')
        response = client.get("/device")
        assert response.status_code == 200

    except Exception as e:
        if response_status_code == 404:
            assert str(e) == "Mock using model example but no example found for the API"
        elif response_status_code == 409:
            assert str(e) == "Mock status code not defined in API declaration"
        else:
            assert False



"""
Module containing the Device model definition.
"""

from typing import ClassVar
from datetime import date
from pydantic import BaseModel, Field, EmailStr


class Coordinate(BaseModel):
    """
    Represents a geographical coordinate with latitude and longitude.

    Attributes:
    - latitude (float): The latitude in decimal degrees, must be between -90 and 90.
    - longitude (float): The longitude in decimal degrees, must be between -180 and 180.
    """
    latitude: float = Field(default=...,
                            description="Latitude in decimal degrees. "
                                        "Must be between -90 and 90.")
    longitude: float = Field(default=...,
                             description="Longitude in decimal degrees. "
                                         "Must be between -180 and 180.")

    model_config = {
        "json_schema_extra": {
            "example": {
                "latitude": 35.6582,
                "longitude": 139.8752
            }
        }
    }


class Device(BaseModel):
    """
    Represents a device with unique identifiers and location information.

    Attributes:
    - device_uuid (str): The UUID of the device. It should start with the prefix (DEV),
      followed by a single uppercase letter and six digits (e.g., DEVX000001).
    - localisation (Coordinate): Geographical location of the device as coordinates.
    - deployment_date (date): The deployment date of the device in the format YYYY-MM-DD.
    - owner (EmailStr): The email address of the owner of the device.
    """
    UUID_REGEX_PATTERN: ClassVar[str] = r'^DEV[A-Z]\d{6}$'

    device_uuid: str = Field(default=...,
                             description="The UUID of the device. "
                                         "It should start with the prefix (DEV), "
                                         "a single uppercase letter, "
                                         "and six digits (e.g., DEVX000001)")

    localisation: Coordinate = Field(default=None,
                                     description="The geographical location of the device "
                                                 "as latitude and longitude coordinates")

    deployment_date: date = Field(default=None,
                                  description="The deployment date of the device "
                                              "in the format YYYY-MM-DD")
    owner: EmailStr = Field(default=...,
                            description="The email address of the owner of the device")

    model_config = {
        "json_schema_extra": {
            "example": {
                "device_uuid": "DEVX000001",
                "localisation": {
                    "latitude": 35.6582,
                    "longitude": 139.8752
                },
                "deployment_date": "2024-03-14",
                "owner": "tanguy.pemeja@gmail.com"
            }
        }
    }


class DeviceNotFoundResponse(BaseModel):
    """
    Represents a response indicating that the device was not found.
    """
    message: str = "Device not found"


class DeviceAlreadyExists(BaseModel):
    """
    Represents a response indicating that the device already exists.
    """
    message: str = "Device already exists"



from fastapi import FastAPI, status
from fastapi.testclient import TestClient
from typing import Any

from fastmock.middleware import FastMockMiddleware
from tests.device import Device, DeviceNotFoundResponse


class CustomTestClient(TestClient):
    def __init__(self, app: FastAPI, headers: dict[str, Any] = None):
        super().__init__(app)
        self._custom_headers = headers or {}

    def request(self, method: str, url: str, **kwargs):
        headers = kwargs.get('headers', {})
        if headers is None:
            headers = {}
        # Convert all header values to strings
        headers.update({key: str(value) for key, value in self._custom_headers.items()})
        kwargs['headers'] = headers
        return super().request(method, url, **kwargs)


def get_client(**kwargs) -> TestClient:
    # Recreate the app to reset the middleware for each test
    test_app = FastAPI()
    test_app.add_middleware(FastMockMiddleware)

    @test_app.get("/list",
                  status_code=status.HTTP_200_OK,
                  responses={
                      status.HTTP_200_OK: {"model": list[Device]},
                      status.HTTP_404_NOT_FOUND: {"model": DeviceNotFoundResponse}
                  })
    async def get_device_list():
        return {"msg": "Device list"}

    @test_app.get("/dict",
                  status_code=status.HTTP_200_OK,
                  responses={
                      status.HTTP_200_OK: {"model": dict[str, Device]},
                      status.HTTP_404_NOT_FOUND: {"model": DeviceNotFoundResponse}
                  })
    async def get_device():
        return {"msg": "Device dict"}

    @test_app.get("/device",
                  status_code=status.HTTP_200_OK,
                  responses={
                      status.HTTP_200_OK: {"model": Device},
                      status.HTTP_404_NOT_FOUND: {"model": DeviceNotFoundResponse}
                  })
    async def get_device():
        return {"msg": "Device"}

    @test_app.get("/str",
                  status_code=status.HTTP_200_OK,
                  responses={
                      status.HTTP_200_OK: {"model": str},
                      status.HTTP_404_NOT_FOUND: {"model": DeviceNotFoundResponse}
                  })
    async def get_str():
        return {"msg": "Str"}

    @test_app.get("/int",
                  status_code=status.HTTP_200_OK,
                  responses={
                      status.HTTP_200_OK: {"model": int},
                      status.HTTP_404_NOT_FOUND: {"model": DeviceNotFoundResponse}
                  })
    async def get_int():
        return {"msg": "Int"}

    custom_headers = {f"X-FASTMOCK-{key.upper().replace('-', '_')}": value for key, value in kwargs.items()}

    return CustomTestClient(test_app, headers=custom_headers)



import pytest
from tests.clients import (decorator_client, decorator_initialization_client,
                           middleware_client, header_client)


@pytest.mark.parametrize("get_client", [
    middleware_client.get_client,
    decorator_client.get_client,
    decorator_initialization_client.get_client,
    header_client.get_client
])
@pytest.mark.parametrize("response_type", ["example", "default", "generated"])
def test_fastmock_model_response_value(get_client, response_type: str):
    client = get_client(type=response_type)
    response = client.get("/device")
    assert response.status_code == 200

    match response_type:
        case "example":
            assert response.json() == {
                "device_uuid": "DEVX000001",
                "localisation": {
                    "latitude": 35.6582,
                    "longitude": 139.8752
                },
                "deployment_date": "2024-03-14",
                "owner": "tanguy.pemeja@gmail.com"
            }
        case "default":
            assert response.json() != {
                "device_uuid": "DEVX000001",
                "localisation": {
                    "latitude": 35.6582,
                    "longitude": 139.8752
                },
                "deployment_date": "2024-03-14",
                "owner": "tanguy.pemeja@gmail.com"
            }
            assert list(response.json().keys()) == ['device_uuid', 'localisation', 'deployment_date', 'owner']
            assert response.json()["localisation"] is None and response.json()["deployment_date"] is None

        case "generated":
            assert response.json() != {
                "device_uuid": "DEVX000001",
                "localisation": {
                    "latitude": 35.6582,
                    "longitude": 139.8752
                },
                "deployment_date": "2024-03-14",
                "owner": "tanguy.pemeja@gmail.com"
            }
            assert list(response.json().keys()) == ['device_uuid', 'localisation', 'deployment_date', 'owner']
            assert (response.json()["localisation"] is not None
                    and response.json()["deployment_date"] is not None)
        case _:
            assert False


@pytest.mark.parametrize("get_client", [
    middleware_client.get_client,
    decorator_client.get_client,
    decorator_initialization_client.get_client,
    header_client.get_client
])
@pytest.mark.parametrize("response_type", ["example", "default", "generated"])
def test_fastmock_str_response_value(get_client, response_type: str):
    try:
        client = get_client(type=response_type)
        response = client.get("/str")
        assert response.status_code == 200

        match response_type:
            case "default":
                assert type(response.json()) == str
            case "generated":
                assert type(response.json()) == str
            case _:
                assert False

    except Exception as e:
        match response_type:
            case "example":
                assert str(e) == "Mock using model example but no example found for the API"
            case _:
                assert False
